/******************************************************************************
 * Copyright 2004, 2005 Alex Karev.
 *
 * This file is part of gtkpivo library.
 *
 * gtkpivo library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation version 2.1 of the License.
 *
 * gtkpivo library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with gtkpivo library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ******************************************************************************/
/* vim: set ts=4 sw=4 : */
#include <string.h>

#include "gtkiofield.h"

#define LABEL_FONT "Sans Bold 9"
#define VAL_FONT "Sans 7"

#define VAL_NAME "1 KVA QUA 1310"


#define IOFIELD_DEFAULT_WIDTH	40
#define IOFIELD_DEFAULT_HEIGHT	20

#define DEF_BORDER_WIDTH 6

static GtkWidgetClass *parent_class = NULL;


static void gtk_iofield_class_init(GtkIOFieldClass *class);
static void gtk_iofield_init(GtkIOField *iof);
static void gtk_iofield_destroy(GtkObject *object);
static void gtk_iofield_realize(GtkWidget *widget);
static void gtk_iofield_size_request(GtkWidget      *widget,
				      GtkRequisition *requisition);
static void gtk_iofield_size_allocate(GtkWidget     *widget,
					   GtkAllocation *allocation);
static gboolean gtk_iofield_expose(GtkWidget      *widget,
				GdkEventExpose *event );

static gboolean draw_iofield_dynamic(GtkIOField *iof);
static gboolean draw_iofield_static(GtkIOField *iof);

static void adj_val_changed(GtkAdjustment *adj, gpointer arg);
static void adj_changed(GtkAdjustment *adj, gpointer arg);
static  void remove_adj(GtkIOField *iof);

static void
gtk_iofield_class_init(GtkIOFieldClass *class)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;
	
	object_class = (GtkObjectClass*) class;
	widget_class = (GtkWidgetClass*) class;

	parent_class = gtk_type_class (gtk_widget_get_type ());

	object_class->destroy = gtk_iofield_destroy;
	
	widget_class->size_request		= gtk_iofield_size_request;
	widget_class->realize			= gtk_iofield_realize;
	widget_class->size_allocate		= gtk_iofield_size_allocate;
	widget_class->expose_event		= gtk_iofield_expose;
}

static void
gtk_iofield_init(GtkIOField *iof)
{
	iof->font_label = pango_font_description_from_string(LABEL_FONT);
	iof->font_val = pango_font_description_from_string(VAL_FONT);

	iof->adj	= NULL;
	iof->label	= NULL;
}


static void
gtk_iofield_destroy(GtkObject *object)
{
	GtkIOField *iof;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GTK_IS_IOFIELD (object));

	iof = GTK_IOFIELD (object);
	
	if (iof->label != NULL)
		g_free(iof->label);
	
	iof->label = NULL;

	if (iof->adj != NULL) 
		remove_adj(iof);

	iof->adj = NULL;

#if 0
	if (taho->red_gc)
		g_object_unref(GTK_OBJECT(taho->red_gc));

	if (taho->green_gc)
		g_object_unref(GTK_OBJECT(taho->green_gc));

	if (taho->yellow_gc)
		g_object_unref(GTK_OBJECT(taho->yellow_gc));

	if (taho->font_label)
		pango_font_description_free (taho->font_label);
	
	if (taho->font_val)
		pango_font_description_free (taho->font_val);

	if (taho->adj)
		g_object_unref(taho->adj);

#endif

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gtk_iofield_realize(GtkWidget *widget)
{
	GtkIOField *iof;
	GdkWindowAttr attributes;
	gint attributes_mask;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_IOFIELD(widget));

	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	iof = GTK_IOFIELD (widget);

	attributes.x		= widget->allocation.x;
	attributes.y		= widget->allocation.y;
	attributes.width	= widget->allocation.width;
	attributes.height	= widget->allocation.height;

	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.event_mask = gtk_widget_get_events (widget) | 
		GDK_EXPOSURE_MASK /*| GDK_STRUCTURE_MASK */;

	attributes.visual = gtk_widget_get_visual (widget);
	attributes.colormap = gtk_widget_get_colormap (widget);

   	attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
	
	widget->window = gdk_window_new( gtk_widget_get_parent_window(widget), 
			&attributes, 
			attributes_mask);
	gdk_window_set_user_data(widget->window, iof);

	gtk_widget_modify_bg(widget, GTK_STATE_NORMAL, 
		&widget->style->white );
	gtk_widget_modify_fg(widget, GTK_STATE_ACTIVE, 
		&widget->style->black );
	
	widget->style = gtk_style_attach(widget->style, widget->window);
	gtk_style_set_background(widget->style, widget->window, GTK_STATE_NORMAL);


}

static void
gtk_iofield_size_allocate(GtkWidget     *widget,
					   GtkAllocation *allocation)
{
	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_IOFIELD (widget));
	g_return_if_fail (allocation != NULL);

	widget->allocation = *allocation;
	(* GTK_WIDGET_CLASS (parent_class)->size_allocate) (widget, allocation);
	
	if (GTK_WIDGET_REALIZED(widget)) {
		gdk_window_move_resize(widget->window,
				allocation->x, allocation->y,
				allocation->width, allocation->height);

	}

}


static void 
gtk_iofield_size_request(GtkWidget      *widget,
				      GtkRequisition *requisition)
{
	requisition->width		= IOFIELD_DEFAULT_WIDTH;
	requisition->height		= IOFIELD_DEFAULT_HEIGHT;
}

static gboolean
gtk_iofield_expose(GtkWidget      *widget,
				GdkEventExpose *event )
{
	GtkIOField *iof = GTK_IOFIELD(widget);

	gint width, height;
	
    width  = widget->allocation.width;
    height = widget->allocation.height;

	if (event->count > 0)
		return (FALSE);

	(void) draw_iofield_dynamic(iof);

	gtk_paint_shadow( widget->style, widget->window,
			widget->state, GTK_SHADOW_ETCHED_OUT,
			NULL /* &event->area */, widget,  "iof",
			// x, y,
			0, 0,
			width, height );

	return (FALSE);
}


static gboolean
draw_iofield_dynamic(GtkIOField *iof)
{
	gchar label[32];
	gint 
		dx, dy;

	g_snprintf(label, sizeof label, "%3.2f", iof->adj->value );

	iof->layout_val = gtk_widget_create_pango_layout(GTK_WIDGET(iof), label);
	pango_layout_set_font_description (iof->layout_val, iof->font_val); 
	pango_layout_get_pixel_size(iof->layout_val, &dx, &dy);

	dx = ( GTK_WIDGET(iof)->allocation.width  - dx ) / 2;
	dy = ( GTK_WIDGET(iof)->allocation.height - dy ) / 2;

	gdk_draw_layout( GTK_WIDGET(iof)->window,
				GTK_WIDGET(iof)->style->black_gc,
				dx , dy,
				iof->layout_val );
	
	g_object_unref(iof->layout_val);

	return (TRUE);
}

static gboolean
draw_iofield_static(GtkIOField *iof)
{
	g_return_val_if_fail(iof != NULL, FALSE);

		
	return (TRUE);
}

static void
adj_changed(GtkAdjustment *adj, gpointer arg)
{
	GtkIOField *iof = (GtkIOField *) arg;

	g_return_if_fail(iof != NULL);
}

static void
adj_val_changed(GtkAdjustment *adj, gpointer arg)
{

	GtkIOField *iof = (GtkIOField *) arg;
	
	g_assert(iof != NULL);
	gtk_widget_queue_draw(GTK_WIDGET(iof));
}

static  void
remove_adj(GtkIOField *iof)
{
	g_return_if_fail(iof != NULL);
	g_return_if_fail(iof->adj != NULL);
	
	g_signal_handlers_disconnect_by_func( G_OBJECT(iof->adj),
			G_CALLBACK(adj_changed), 
			iof);
	g_signal_handlers_disconnect_by_func( G_OBJECT(iof->adj),
			G_CALLBACK(adj_val_changed), 
			iof);

	g_object_unref(iof->adj);
}


/** 
 *	
 *	Public functions
 * 
 */
GtkType
gtk_iofield_get_type(void)
{
	static GtkType iof_type = 0;

	if (!iof_type)
	{
		static const GtkTypeInfo iof_info = 
		{
			"GtkIOField",
			sizeof (GtkIOField),
			sizeof (GtkIOFieldClass),
			(GtkClassInitFunc) gtk_iofield_class_init,
			(GtkObjectInitFunc) gtk_iofield_init,
			/* reserved_1 */ NULL,
			/* reserved_1 */ NULL,
			(GtkClassInitFunc) NULL
		};

		iof_type = gtk_type_unique (GTK_TYPE_WIDGET, &iof_info);
	}
	
	return (iof_type);
}

GtkWidget* 
gtk_iofield_new(GtkAdjustment *adj, const gchar *label )
{
	GtkIOField *iof;
		
	iof = gtk_type_new(gtk_iofield_get_type());

	/** Adjustment setup */
	gtk_iofield_set_adjustment(iof, adj);

#if 0
	/** Static labels setup */
	setup_static_text(iof);
#endif
	
	/** Creating label */
	if (label == NULL) {
		iof->label = g_strndup(VAL_NAME, strlen(VAL_NAME));
	} else {
		iof->label = g_strndup(label, 32);
	}

	return GTK_WIDGET(iof);
}

void
gtk_iofield_set_adjustment(GtkIOField *iof, GtkAdjustment *adj)
{
	g_return_if_fail( iof != NULL );
	g_return_if_fail( GTK_IS_IOFIELD(iof) );

	if (adj) 
		g_return_if_fail(GTK_IS_ADJUSTMENT(adj));
	else 
		adj = (GtkAdjustment *) gtk_adjustment_new(0, 0., 999., 0.001, 1, 10);

	
	if (iof->adj != adj) {
		if (iof->adj != NULL)
			remove_adj(iof);


		iof->adj = adj;
		if (adj) {
			g_object_ref(adj);
			gtk_object_sink(GTK_OBJECT(adj));

			g_signal_connect(G_OBJECT(adj), "changed", G_CALLBACK(adj_changed),
				iof);
			g_signal_connect(G_OBJECT(adj), "value-changed", G_CALLBACK(adj_val_changed),
				iof);
		}
	}
}

