/******************************************************************************
 * Copyright 2004, 2005 Alex Karev.
 *
 * This file is part of gtkpivo library.
 *
 * gtkpivo library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation version 2.1 of the License.
 *
 * gtkpivo library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with gtkpivo library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ******************************************************************************/
/* vim: set ts=4 sw=4 : */

#define _USE_MATH_DEFINES
#include <math.h>
#include <string.h>

#include "gtktrend.h"


#define GTK_TREND_SET_LINE_SIZE 3
#define GTK_TREND_LINE_SIZE 2

#define TREND_DEFAULT_WIDTH  660 
#define TREND_DEFAULT_HEIGHT 200 

#define TREND_MARGINE		8 /* pixels */
#define TREND_PLOT_OFFSET   8

#define TREND_DX			2 /* Multiply by * (1 << TREND_DX) */

#define TREND_ADJ_MAX	128
#define TREND_ADJ_MIN 	-128

#define TREND_LABEL "Unit #1 input voltage (V)"

#define LABEL_FONT "Sans bold 8"
#define GRID_FONT "Sans 6"
#define VAL_FONT "Sans 7"

#define BAR_ASPECT_RATIO 6
#define BAR_MARGINE		4

#define VAL_AREA_MARGINE 8

#define GRID_LINE_COLOR1 "Green"
#define GRID_LINE_COLOR2 "DarkGreen"
#define GRID_TICK_LENGTH  BAR_MARGINE	

#define TREND_TICKS_NUMBER 20

#define CLEAR_QUEUE() \
do { \
	int i; \
	\
	/* g_print("CLEAR_QUEUE\n"); */	\
	for (i = 0; i < TREND_SIGNALS; i++) { \
		trend->head[i] = trend->n[i] = trend->tail[i] = 0; \
		memset(trend->history[i], 0, sizeof trend->history[i]); \
	} \
} while(0)

/* FIXME: not tested yet. Is it really needed */
#define GET_BUFFER(b, n, head) ((n == 0) ? 0 : (b)[(n)-- % (TREND_HISTORY)])

static GtkWidgetClass *parent_class = NULL;

static void gtk_trend_class_init(GtkTrendClass *class);
static void gtk_trend_init(GtkTrend *trend);
static void gtk_trend_destroy (GtkObject *object);
static void gtk_trend_realize(GtkWidget *widget);
static void gtk_trend_size_request(GtkWidget      *widget,
				      GtkRequisition *requisition);
static void gtk_trend_size_allocate(GtkWidget     *widget,
					   GtkAllocation *allocation);
static gboolean gtk_trend_expose(GtkWidget      *widget,
				GdkEventExpose *event );
static gboolean gtk_trend_button_press( GtkWidget *widget,
		       GdkEventButton *event );
static gboolean gtk_trend_button_release( GtkWidget *widget,
			   GdkEventButton *event );
static gboolean gtk_trend_motion_notify( GtkWidget *widget,
               GdkEventMotion *event );
static void draw_trend_static(GtkTrend *t);
static void draw_values_selection(GtkTrend *t, gint x, gint y);
static void draw_trend_dynamic(GtkTrend *t);
/* static void draw_trend_text_static(GtkTrend *t); */
static void recount_deometry_params(GtkTrend *t);
static void put_queue(GtkTrend *t, gint index, gfloat val);
static void adj_changed(GtkAdjustment *adj, gpointer arg);
static void adj_val_changed(GtkAdjustment *adj, gpointer arg);

static struct adj_signal_arg {
	GtkTrend *t;
	gint i;
} adj_sig_args[ TREND_SIGNALS ];

/* 4 for now */
static const gchar *line_colors[ TREND_SIGNALS ] = {
#if 0
	"light green",
	"grey",
	"light yellow",
	"light blue"
#else
	"DodgerBlue",
	"LightGoldenrod",
	"cyan",
	"orchid2"
#endif
};

static void 
put_queue(GtkTrend *t, gint index, gfloat val)
{
	if (t->n[index] >= TREND_HISTORY) {
		t->head[index] = (t->head[index] + 1) % TREND_HISTORY;
	} else {
		t->n[index]++;
	}

	t->tail[index] = (t->tail[index] + 1) % TREND_HISTORY;
	t->history[index][t->tail[index]] = val;
}

/*
 * 
 * Adjustments section
 *
 * TODO: Update plot even new value is the same as prev.
 *
 */
static void
adj_changed(GtkAdjustment *adj, gpointer arg)
{
	struct  adj_signal_arg *adj_arg = (struct adj_signal_arg *) arg;

	/* Optimized, may be not safe
	g_return_if_fail(adj_arg != NULL);
	g_return_if_fail(adj_arg->t != NULL);
	g_return_if_fail(adj_arg->i < TREND_SIGNALS);
	*/
	
#if 0
	g_print("changed\n");
#endif

	gtk_widget_queue_draw(GTK_WIDGET(adj_arg->t));
}

static void
adj_val_changed(GtkAdjustment *adj, gpointer arg)
{
	struct  adj_signal_arg *adj_arg = (struct adj_signal_arg *) arg; 

	/* Optimized, may be not safe
	g_return_if_fail(adj_arg != NULL);
	g_return_if_fail(adj_arg->t != NULL);
	g_return_if_fail(adj_arg->i < TREND_SIGNALS);
	*/

#if 0
	g_print("val_changed\n"); 
#endif
	
#if 1
	put_queue(adj_arg->t, adj_arg->i, adj->value);
	gtk_widget_queue_draw(GTK_WIDGET(adj_arg->t));
#endif

}

void
gtk_trend_update_plot(GtkTrend *t)
{
	gtk_adjustment_value_changed(t->adj[0]);
	gtk_adjustment_value_changed(t->adj[1]);
	gtk_adjustment_value_changed(t->adj[2]);
	gtk_adjustment_value_changed(t->adj[3]);
}

#if 0
void
gtk_trend_append_values(GtkTrend *t, gfloat val1, gfloat val2, gfloat val3, gfloat val4)
{
	/* Optimized, may be not safe
	g_return_if_fail(t != NULL);
	g_return_if_fail(TREND_SIGNALS <= 4);
	*/
	
	/*
	put_queue(t, 0, val1);
	put_queue(t, 1, val2);
	put_queue(t, 2, val3);
	put_queue(t, 3, val4);
	

	gtk_adjustment_set_value(t->adj[0], val1);
	gtk_adjustment_set_value(t->adj[1], val1);
	gtk_adjustment_set_value(t->adj[2], val1);
	gtk_adjustment_set_value(t->adj[3], val1);
	*/
	
	gtk_widget_queue_draw(GTK_WIDGET(t));
	
}
#endif

GtkAdjustment*	
gtk_trend_get_adjustment(GtkTrend *t, gint index)
{
	g_return_val_if_fail(t != NULL, NULL);
	g_return_val_if_fail(GTK_IS_TREND(t), NULL);
	g_return_val_if_fail(index < TREND_SIGNALS, NULL);

	return (t->adj[ index ]);
}

GtkType
gtk_trend_get_type(void)
{
	static GtkType trend_type = 0;

	if (!trend_type)
	{
		static const GtkTypeInfo trend_info = 
		{
			"GtkTrend",
			sizeof (GtkTrend),
			sizeof (GtkTrendClass),
			(GtkClassInitFunc) gtk_trend_class_init,
			(GtkObjectInitFunc) gtk_trend_init,
			/* reserved_1 */ NULL,
			/* reserved_1 */ NULL,
			(GtkClassInitFunc) NULL
		};

		trend_type = gtk_type_unique (GTK_TYPE_WIDGET, &trend_info);
	}
	
	return (trend_type);
}

static void
gtk_trend_class_init(GtkTrendClass *class)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;
	
	object_class = (GtkObjectClass*) class;
	widget_class = (GtkWidgetClass*) class;

	parent_class = gtk_type_class (gtk_widget_get_type ());

	object_class->destroy = gtk_trend_destroy;
	
	widget_class->realize		= gtk_trend_realize;
	widget_class->expose_event	= gtk_trend_expose;
	widget_class->size_request	= gtk_trend_size_request;
	widget_class->size_allocate	= gtk_trend_size_allocate;
	
	widget_class->button_press_event = gtk_trend_button_press;
	widget_class->button_release_event = gtk_trend_button_release;
	widget_class->motion_notify_event = gtk_trend_motion_notify;
}

static void
gtk_trend_init(GtkTrend *trend)
{
	/** Initializing of  constant values */
	trend->trend_left_margine =
	trend->trend_right_margine =
   	trend->trend_top_margine = 
	trend->trend_buttom_margine = TREND_MARGINE;

	trend->plot_top_offset =
	trend->plot_buttom_offset =
	trend->plot_left_offset =
	trend->plot_right_offset = TREND_PLOT_OFFSET;
	
	trend->bar_margine = BAR_MARGINE;
	trend->values_area_margine = VAL_AREA_MARGINE;

	trend->xc = trend->xend = trend->yc = 0;

	trend->button = 0;

	trend->sel_x = 0, trend->sel_y = 0;

	/** Buffer init */
	CLEAR_QUEUE();
	recount_deometry_params(trend);

	/** Fonts setup */
	trend->font_grid_labels = pango_font_description_from_string (GRID_FONT);
	trend->font_trend_label = pango_font_description_from_string (LABEL_FONT);
	trend->font_val = pango_font_description_from_string (VAL_FONT);

	trend->label = NULL;
	trend->draw_value_selection_flag = 0;

	memset(trend->points, 0, TREND_HISTORY);

#if 0
	{
		int i, j;

		for(i = 0 ; i < TREND_SIGNALS; i++)
			for(j = 0 ; j < TREND_HISTORY;j++) {
				put_queue(trend, i, 20 * (i + 1) * sin((j + i) * M_PI / 180.));
			}
	}
#endif

}

/*
 *  FIXME: Rewrite trend lines management and initialization
 *
 */
GtkWidget*
gtk_trend_new( gchar *label,  
	GtkAdjustment *adj1,
	GtkAdjustment *adj2,
	GtkAdjustment *adj3,
	GtkAdjustment *adj4	)
{
	GtkTrend *trend;
		
	trend = gtk_type_new(gtk_trend_get_type());

	trend->label = (NULL == label) ? g_strdup(TREND_LABEL) : g_strdup(label);

	/* Adjustments setup for 4 signals */
	adj_sig_args[0].t = trend;
	adj_sig_args[0].i = 0;
	trend->adj[0] = (NULL == adj1) ? (GtkAdjustment *) gtk_adjustment_new(0, TREND_ADJ_MIN, TREND_ADJ_MAX, 0.1, 1, 1) : adj1;
#if 1
	g_signal_connect(G_OBJECT(trend->adj[0]), "changed", G_CALLBACK(adj_changed),
		   	&adj_sig_args[0]);
	g_signal_connect(G_OBJECT(trend->adj[0]), "value-changed", G_CALLBACK(adj_val_changed),
		   	&adj_sig_args[0]);
#endif
	
	adj_sig_args[1].t = trend;
	adj_sig_args[1].i = 1;
	trend->adj[1] = (NULL == adj2) ? (GtkAdjustment *) gtk_adjustment_new(0, 20, 128, 0.1, 1, 1) : adj2;
#if 1
	g_signal_connect(G_OBJECT(trend->adj[1]), "changed", G_CALLBACK(adj_changed),
		   	&adj_sig_args[1]);
	g_signal_connect(G_OBJECT(trend->adj[1]), "value-changed", G_CALLBACK(adj_val_changed),
		   	&adj_sig_args[1]);
#endif
	
	
	adj_sig_args[2].t = trend;
	adj_sig_args[2].i = 2;
	trend->adj[2] = (NULL == adj3) ? (GtkAdjustment *) gtk_adjustment_new(-100, -500, -20, 0.1, 1, 1) : adj3;
#if 1
	g_signal_connect(G_OBJECT(trend->adj[2]), "changed", G_CALLBACK(adj_changed),
		   	&adj_sig_args[2]);
	g_signal_connect(G_OBJECT(trend->adj[2]), "value-changed", G_CALLBACK(adj_val_changed),
		   	&adj_sig_args[2]);
#endif
	
	adj_sig_args[3].t = trend;
	adj_sig_args[3].i = 3;
	trend->adj[3] = (NULL == adj4) ? (GtkAdjustment *) gtk_adjustment_new(0, 0, 128, 0.1, 1, 1) : adj4;
#if 1
	g_signal_connect(G_OBJECT(trend->adj[3]), "changed", G_CALLBACK(adj_changed),
		   	&adj_sig_args[3]);
	g_signal_connect(G_OBJECT(trend->adj[3]), "value-changed", G_CALLBACK(adj_val_changed),
		   	&adj_sig_args[3]);
#endif

	return GTK_WIDGET(trend);
}

#if 0
void
gtk_trend_append(GtkTrend *t, gdouble val1, gdouble val2, gdouble val3, gdouble val4 )
{
	g_return_if_fail(t != NULL);
	g_return_if_fail(TREND_SIGNALS >= 4);


	put_queue(t, 0, val1);
	put_queue(t, 1, val2);
	put_queue(t, 2, val3);
	put_queue(t, 3, val4);
	/*
	PUT_BUFFER(t->history[0], t->n, t->head, val1 );
	PUT_BUFFER(t->history[1], t->n, t->head, val2 );

	PUT_BUFFER(t->history[2], t->n, t->head, val3 );
	PUT_BUFFER(t->history[3], t->n, t->head, val4 );
	*/

	gtk_widget_queue_draw(GTK_WIDGET(t));
}
#endif

static void
gtk_trend_destroy (GtkObject *object)
{
	GtkTrend *trend;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GTK_IS_TREND (object));

	trend = GTK_TREND (object);
	if (trend->label != NULL)
		g_free(trend->label);

	trend->label = NULL;

#if 0
	if(trend->grid_line_gc1)
		g_object_unref(trend->grid_line_gc1);

	if(trend->grid_line_gc2)
		g_object_unref(trend->grid_line_gc2);

	if(trend->sel_line_gc)
		g_object_unref(trend->sel_line_gc);
#endif

#if 0
	if (trend->bgpix) 
		g_object_unref(trend->bgpix);
#endif
#if 0
	if (trend->font_grid_labels)
		pango_font_description_free (trend->font_grid_labels);

	if (trend->font_trend_label)
		pango_font_description_free (trend->font_trend_label);
	
	if (trend->font_val)
		pango_font_description_free (trend->font_val);
#endif
	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gtk_trend_realize(GtkWidget *widget)
{
	GtkTrend *trend;
	GdkColor c;
	GdkWindowAttr attributes;
	gint attributes_mask, i;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_TREND (widget));

	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	trend = GTK_TREND (widget);

	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;

	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.event_mask = gtk_widget_get_events (widget) | 
		GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK | 
		GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK;

	attributes.visual = gtk_widget_get_visual (widget);
	attributes.colormap = gtk_widget_get_colormap (widget);

   	attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
	
	widget->window = gdk_window_new(gtk_widget_get_parent_window(widget), &attributes, attributes_mask);
	widget->style = gtk_style_attach(widget->style, widget->window);
	
	gdk_window_set_user_data(widget->window, widget);
	gtk_style_set_background(widget->style, widget->window, GTK_STATE_ACTIVE);
	
	recount_deometry_params(trend);

	/* Init GCs */
	trend->grid_line_gc1 = gdk_gc_new(GTK_WIDGET(trend)->window);
	gdk_color_parse(GRID_LINE_COLOR1, &c);
	gdk_gc_set_rgb_fg_color(trend->grid_line_gc1, &c);
	
	trend->grid_line_gc2 = gdk_gc_new(GTK_WIDGET(trend)->window);
	gdk_color_parse(GRID_LINE_COLOR2, &c);
	gdk_gc_set_rgb_fg_color(trend->grid_line_gc2, &c);

	trend->sel_line_gc = gdk_gc_new(GTK_WIDGET(trend)->window);
	gdk_color_parse(GRID_LINE_COLOR1, &c);
	gdk_gc_set_rgb_fg_color(trend->sel_line_gc, &c);
	gdk_gc_set_line_attributes(trend->sel_line_gc,
			GTK_TREND_SET_LINE_SIZE, 
			GDK_LINE_SOLID,
			GDK_CAP_ROUND,
			GDK_JOIN_ROUND);

	/* Line GCs */
	for(i = 0; i < TREND_SIGNALS; i++) {
		trend->line_gc[i] =  gdk_gc_new(GTK_WIDGET(trend)->window);
		gdk_color_parse(line_colors[i], &c);
		gdk_gc_set_rgb_fg_color(trend->line_gc[i], &c);
		gdk_gc_set_line_attributes(trend->line_gc[i],
			GTK_TREND_LINE_SIZE,  /* Line size */
			GDK_LINE_SOLID,
			GDK_CAP_ROUND,
			GDK_JOIN_ROUND);
	}

	if (trend->bgpix != NULL) {
		gdk_window_set_back_pixmap(widget->window, NULL, FALSE);
		g_object_unref(trend->bgpix);
	}

#if 1
	trend->bgpix = gdk_pixmap_new(widget->window, 
			widget->allocation.width, widget->allocation.height,
			-1);
#else
	trend->bgpix = gdk_pixmap_new(NULL, 
			widget->allocation.width, widget->allocation.height,
			1);
#endif

#if 0
	gdk_draw_drawable (widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			trend->bgpix,
			/* Only copy the area that was exposed. */
			0, 0,
			0, 0,
			widget->allocation.width, widget->allocation.height);
#endif

	draw_trend_static(trend);
	gdk_window_set_back_pixmap(widget->window, trend->bgpix, FALSE);
	draw_trend_dynamic(trend);
}

static void 
gtk_trend_size_request(GtkWidget      *widget,
				      GtkRequisition *requisition)
{
	requisition->width = TREND_DEFAULT_WIDTH;
	requisition->height = TREND_DEFAULT_HEIGHT;
}

static void
gtk_trend_size_allocate(GtkWidget     *widget,
					   GtkAllocation *allocation)
{
	GtkTrend *trend = GTK_TREND(widget);

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_TREND (widget));
	g_return_if_fail (allocation != NULL);

	widget->allocation = *allocation;
	
	if (GTK_WIDGET_REALIZED(widget)) {
		trend = GTK_TREND (widget);
		gdk_window_move_resize(widget->window,
				allocation->x, allocation->y,
				allocation->width, allocation->height);
	}

	recount_deometry_params(trend);

	if (widget->window == NULL)
		return;

	if (trend->bgpix != NULL)  {
		gdk_window_set_back_pixmap(widget->window, NULL, FALSE);
		g_object_unref(trend->bgpix);
	}
	
#if 1
	trend->bgpix = gdk_pixmap_new(widget->window, 
			widget->allocation.width, widget->allocation.height,
			-1);
#else
	trend->bgpix = gdk_pixmap_new(NULL, 
			widget->allocation.width, widget->allocation.height,
			1);
#endif

	draw_trend_static(trend);
	gdk_window_set_back_pixmap(widget->window, trend->bgpix, FALSE);
	draw_trend_dynamic(trend);
}

static gboolean 
gtk_trend_button_press( GtkWidget *widget,
		       GdkEventButton *event )
{
	GtkTrend *trend;
	gint left_border, right_border, top_border, buttom_border;
	
	g_return_val_if_fail(widget != NULL, FALSE);
	g_return_val_if_fail(GTK_IS_TREND(widget), FALSE);
	g_return_val_if_fail(event != NULL, FALSE);

	trend = GTK_TREND(widget);
	
#if 0
	g_print("pressed\n");
#endif

	left_border = trend->plot_left_offset + trend->trend_left_margine;
	right_border = left_border + trend->plot_signals_area_width;

	top_border = trend->plot_top_offset + trend->trend_top_margine + trend->main_label_area_height;
	buttom_border = top_border + trend->plot_area_height;
	
	/** If no buttons pressed yet and we clicked inside plot aread, then... */
	if (!trend->button && 
			(event->x > left_border && event->x < right_border) &&
			(event->y > top_border && event->y < buttom_border)) {

		trend->draw_value_selection_flag = 1;
		
		/** all input goes only to this widget */
		gtk_grab_add(widget);

#if 0
		g_print("kva\n");
#endif

		trend->button = event->button;

		trend->sel_x = event->x, trend->sel_y = event->y;
		trend->draw_value_selection_flag = 1;
		gtk_widget_queue_draw(widget);
	}
	return (FALSE);
}

static gboolean 
gtk_trend_button_release( GtkWidget *widget,
			   GdkEventButton *event )
{
	GtkTrend *trend;
	
	g_return_val_if_fail(widget != NULL, FALSE);
	g_return_val_if_fail(GTK_IS_TREND(widget), FALSE);
	g_return_val_if_fail(event != NULL, FALSE);

	trend = GTK_TREND(widget);
	
	if (event->button == trend->button) {
#if 0
		g_print("release\n");
#endif
		GTK_TREND(widget)->button = 0;
		trend->draw_value_selection_flag = 0;
		gtk_widget_queue_draw(widget);
		gtk_grab_remove(widget);
	}
	
	return (FALSE);
}

static gboolean 
gtk_trend_motion_notify( GtkWidget *widget,
               GdkEventMotion *event )
{
	GtkTrend *trend;
	GdkModifierType mods;
	gint x, y, mask;
  
	g_return_val_if_fail(widget != NULL, FALSE);
	g_return_val_if_fail(GTK_IS_TREND(widget), FALSE);
	g_return_val_if_fail(event != NULL, FALSE);

	trend = GTK_TREND(widget);

	if (trend->button != 0) { /** some mouse button pressed */
		x = event->x;
	    y = event->y;

#if 0 /** for study/test */
		if (event->is_hint) {
			g_print("is_hint\n");
			gdk_window_get_pointer (widget->window, &x, &y, &mods);
		} else if (event->window != widget->window) {
			g_print("event->window != widget->window\n");
			gdk_window_get_pointer (widget->window, &x, &y, &mods);
		}
#endif 


		if (event->is_hint || (event->window != widget->window)) {
			gdk_window_get_pointer (widget->window, &x, &y, &mods);
		}

		
		switch (trend->button) {
			case 1:
				mask = GDK_BUTTON1_MASK;
				break;
			case 2:
				mask = GDK_BUTTON2_MASK;
				break;
			case 3:
				mask = GDK_BUTTON3_MASK;
				break;
			default:
				mask = 0;
				break;
		}

		if (mods & mask) {
			/*
			g_print("motion: x = %d y = %d\n", x, y);
			*/
			trend->sel_x = x, trend->sel_y = y;
			trend->draw_value_selection_flag = 1;
			gtk_widget_queue_draw(widget);
		}
	}

	return (FALSE);
}

static gboolean
gtk_trend_expose(GtkWidget      *widget,
				GdkEventExpose *event )
{
	GtkTrend *trend = GTK_TREND(widget);
	
	/*
	g_print("expose event\n");
	*/

	/* Removed for optimization */
	/*
	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_TREND (widget), FALSE);
	*/

	if (event->count > 0) /* Why? */
	    return (FALSE);

#if 0
	gdk_draw_drawable (widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			trend->bgpix,
			/* Only copy the area that was exposed. */
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height);
#endif
	
	draw_trend_dynamic(trend);
	if (trend->draw_value_selection_flag) 
		draw_values_selection(trend, trend->sel_x, trend->sel_y );

	return (FALSE);
}

/* Draw routines */
void
draw_trend_static(GtkTrend *t)
{
	gint width, height, tick_length;
	gint l, i, 
		dx, dy, 
		y1, y2,
		x1, x2;
	gchar sval[32];
	GdkPoint m[5]; /* grid points */

	recount_deometry_params(t);

	/** Fill background */
	gdk_drawable_get_size(GDK_DRAWABLE(t->bgpix), &width, &height);
	gdk_draw_rectangle(t->bgpix,
			GTK_WIDGET(t)->style->black_gc,
			TRUE,
			0, 0,
			width, height);

	/** Draw main label */
	t->layout = gtk_widget_create_pango_layout(GTK_WIDGET(t), t->label);
	pango_layout_set_font_description (t->layout, t->font_trend_label); 
	pango_layout_get_pixel_size(t->layout, &dx, &dy);

	gdk_draw_layout (t->bgpix,
			 GTK_WIDGET(t)->style->white_gc,
			 t->trend_left_margine + (t->main_label_area_width / 2 - dx / 2),  
			 t->trend_top_margine + (t->main_label_area_height / 2 - dy / 2), 
			 t->layout);
	g_object_unref(t->layout);

#if 1
	/**
	 * Draw grid lines
	 */
	/* vertical */
	y1 = t->trend_top_margine + t->plot_top_offset + t->main_label_area_height;
	y2 = y1 + t->plot_area_height;
	
	for( dx = t->xc; dx > t->trend_left_margine + t->plot_left_offset; 
			dx-=(1<<(TREND_DX + 2))) {
		gdk_draw_line(t->bgpix,
			t->grid_line_gc2,
			dx, y1,
			dx, y2);
	}
	
	/* horizontal */
	x1 = t->trend_left_margine + t->plot_left_offset; 
	x2 = x1 + t->plot_area_width - t->bar_area_width;
	for( i = 0; i < TREND_TICKS_NUMBER; i++) {
		gdk_draw_line(t->bgpix,
			t->grid_line_gc2,
			x1, t->yc - (i + 1) * t->plot_area_height / TREND_TICKS_NUMBER ,
			x2, t->yc - (i + 1) * t->plot_area_height / TREND_TICKS_NUMBER);

	}

#endif 

	/* Draw grid border */
	m[0].x = t->trend_left_margine + t->plot_left_offset;
	m[0].y = t->trend_top_margine + t->plot_top_offset + t->main_label_area_height;

	m[1].x = m[0].x;
	m[1].y = m[0].y + t->plot_area_height;
	
	m[2].x = m[1].x + t->plot_area_width - t->bar_area_width;
	m[2].y = m[1].y;

	m[3].x = m[2].x + t->bar_area_width;
	m[3].y = m[2].y;

#if 1
	/* Middle separator */
	m[4].x = m[2].x;
	m[4].y = m[0].y;

	gdk_draw_line(t->bgpix,
			t->grid_line_gc2,
			m[2].x, m[2].y,
			m[4].x, m[4].y);
#endif	


	gdk_draw_lines(t->bgpix,
			t->grid_line_gc1,
			m, G_N_ELEMENTS(m) - 1 );


	
#if 1
	/** Draw bar scales */
	/* 
	 * formula: (2l-1) m + (l-1) b 
	 */
	for(l = 0; l < TREND_SIGNALS;l++) {
		dx = t->xc + ((l + 1)<<1) * t->bar_margine + (l + 1) * t->bar_width;
		gdk_draw_line(t->bgpix,
			t->grid_line_gc1,
			dx, t->yc - t->plot_area_height,
			dx, t->yc);

		/** draw ticks */
		for(i=0; i < TREND_TICKS_NUMBER; i++) {
			
			tick_length = (((i+1) % (TREND_TICKS_NUMBER >> 2)) ? (BAR_MARGINE >> 2): (BAR_MARGINE >> 1));
			dy = t->yc - (i + 1) * t->plot_area_height / TREND_TICKS_NUMBER;
			
			gdk_draw_line(t->bgpix,
				t->grid_line_gc1,
				dx, dy,
				dx - tick_length, dy );
		}

		/** Print scale vertical labels */
		/* Upper */	
		g_snprintf(sval, sizeof sval, "%2.2f", t->adj[l]->upper);
		t->layout = gtk_widget_create_pango_layout(GTK_WIDGET(t), sval);
		pango_layout_set_font_description (t->layout, t->font_grid_labels); 
	   	pango_layout_get_pixel_size (t->layout, &width, &height); 
		
		gdk_draw_layout (t->bgpix,
				t->grid_line_gc1,
				dx - width,  dy - height - 2, 
				t->layout);
		g_object_unref(t->layout);

		/* Lower */
		g_snprintf(sval, sizeof sval, "%2.2f", t->adj[l]->lower);
		t->layout = gtk_widget_create_pango_layout(GTK_WIDGET(t), sval);
		pango_layout_set_font_description (t->layout, t->font_grid_labels); 
	   	pango_layout_get_pixel_size (t->layout, &width, &height); 

		gdk_draw_layout (t->bgpix,
				t->grid_line_gc1,
				dx - width,  t->yc + 2/*+ height*/, 
				t->layout);
		g_object_unref(t->layout);

	}
#endif

}

static void
draw_values_selection(GtkTrend *t, gint x, gint y)
{
	gint y1, y2;
	
	y1 = t->yc - t->plot_area_height - 1 + GTK_TREND_LINE_SIZE ;
	y2 = t->yc + 1 - GTK_TREND_LINE_SIZE;
	
	x = x / (1 << TREND_DX);
	x *= (1 << TREND_DX);

	if ( x < (t->plot_left_offset + t->trend_left_margine) ) {
		x = (t->plot_left_offset + t->trend_left_margine);
	} else if  ( x > t->xc){
		x = t->xc;
	}

	/** First of all draw line itself */
	gdk_draw_line( GTK_WIDGET(t)->window,
				t->sel_line_gc,
				x, y1,
				x, y2);
}

static void 
draw_trend_dynamic(GtkTrend *t)
{
	gfloat dy,k, tmpval;
	gchar sval[32];
	gint l, i, dx, x2, j;
	gint width, height;


	for(l = 0; l < TREND_SIGNALS; l++) {

		if (t->n[l] == 0) continue;

		dy = t->adj[l]->upper - t->adj[l]->lower;
		k = t->plot_area_height / dy;

		/**  Draw signal  history */
		for(i = t->n[l], dx = 0, x2 = 1, j = 0; 
				(i > 0) && (x2 < t->plot_signals_area_width) && (j < TREND_HISTORY);
				i--, dx++, x2 = dx << TREND_DX, j++ ) {

#if 0
			g_print("t->history[%d][%d] = %f\n", l, i, t->history[l][ (t->head[l] + i) % TREND_HISTORY ]);
#endif

			t->points[j].x = t->xc - x2,
			t->points[j].y = 
				t->yc - ( t->history[l][ (t->head[l] + i)     % TREND_HISTORY] - t->adj[l]->lower) * k;


#if 0	/** Old version. Now we just fill up points array here */
			gdk_draw_line(GTK_WIDGET(t)->window,
				t->line_gc[l],
				
				t->xc - x2,
				t->yc - t->history[l][ (t->head[l] + i) % TREND_HISTORY] * k,

				t->xc - (x2 + (1 << TREND_DX)),
				t->yc - t->history[l][ (t->head[l] + i - 1) % TREND_HISTORY] * k
			);
#endif
		}
	
		/** 
		 * Draw all lines in one time, shoud be faster then drow each line, 
		 * but I never test both methods for performance.
		 */
		gdk_draw_lines(GTK_WIDGET(t)->window,
				t->line_gc[l],
				t->points,
				j );

		
		
#if 0 /** Old version. Left here for historical reasons */
		for(i = t->n[l], dx = 0, x2 = ((dx + 1) << TREND_DX); 
			(i > 0) 
			&& (x2 < t->plot_signals_area_width); 
			i--, dx++, x2 = ((dx + 1) << TREND_DX)) {

			gdk_draw_line( GTK_WIDGET(t)->window,
				t->line_gc[l],

				t->xc - (dx << TREND_DX), 
				t->yc - t->history[l][ (t->head[l] + i) % TREND_HISTORY] * k,
				
				t->xc - x2, 
				t->yc - t->history[l][ (t->head[l] + i - 1) % TREND_HISTORY] * k);
		}
#endif

		
		/** Draw bar */
		/* 
		 * formula: (2l-1) m + (l-1) b 
		 */
		tmpval = (t->history[l][ t->tail[l] ] - t->adj[l]->lower)* k;
		gdk_draw_rectangle( GTK_WIDGET(t)->window,
				t->line_gc[l],
				TRUE,
				
				t->xc + ( ((l+1) << 1) - 1) * t->bar_margine + (l) * t->bar_width,
				t->yc - tmpval,
				
				t->bar_width,
				tmpval
			);

		/** Print signal values */
		g_snprintf(sval, sizeof sval, "%2.2f", t->adj[l]->value);
		t->layout = gtk_widget_create_pango_layout(GTK_WIDGET(t), sval);
		pango_layout_set_font_description (t->layout, t->font_val); 
	   	pango_layout_get_pixel_size (t->layout, &width, &height); 


		gdk_draw_layout (GTK_WIDGET(t)->window,
				t->line_gc[l],
				t->values_area_xend - width,
				t->values_area_ystart + (height + 4) * l, 
				t->layout);

#if 0 /* Old version. It's still here for historical reason */
		gdk_draw_layout (GTK_WIDGET(t)->window,
				t->line_gc[l],
				t->xc + (((l+1) << 1) - 1) * t->bar_margine + (l) * t->bar_width, 
				t->yc + 10 + height, 
				t->layout);
#endif

		/*
		gdk_draw_layout (GTK_WIDGET(t)->window,
				t->line_gc[l],
				t->,
				t->,
				t->layout);
				*/


		g_object_unref(t->layout);
		/** End: Print signal values */

	}
}

/* Setup dynamic geometric parameters */
static void 
recount_deometry_params(GtkTrend *trend)
{
#if 0
	trend->trend_buttom_margine = GTK_WIDGET(trend)->allocation.height / 5;
	trend->trend_top_margine =  GTK_WIDGET(trend)->allocation.height / 9;
#endif

	trend->main_label_area_height = 
		(GTK_WIDGET(trend)->allocation.height - 
		 trend->trend_buttom_margine - 
		 trend->trend_top_margine ) / 10;

	trend->main_label_area_width = 
		GTK_WIDGET(trend)->allocation.width - 
	   	trend->trend_left_margine - 
		trend->trend_right_margine;

	trend->values_area_width = ( GTK_WIDGET(trend)->allocation.width - 
		trend->trend_right_margine - trend->trend_right_margine ) / 15;
	trend->values_area_height = 
		GTK_WIDGET(trend)->allocation.height - 
		trend->trend_top_margine - trend->trend_buttom_margine - 
		trend->main_label_area_height;

	trend->values_area_ystart = trend->trend_top_margine + 
	   	trend->main_label_area_height + trend->values_area_margine;
	trend->values_area_yend = trend->values_area_ystart + trend->values_area_height;
	trend->values_area_xstart =  GTK_WIDGET(trend)->allocation.width - trend->trend_right_margine -
		trend->values_area_width - trend->values_area_margine;
	trend->values_area_xend = trend->values_area_xstart + trend->values_area_width;

/*
	trend->trend_right_margine = GTK_WIDGET(trend)->allocation.width / 7;
*/
		
	trend->plot_area_width = GTK_WIDGET(trend)->allocation.width - 
		trend->plot_left_offset - trend->plot_right_offset - 
		trend->trend_left_margine - trend->trend_right_margine - trend->values_area_width;

	trend->plot_area_height = 
		GTK_WIDGET(trend)->allocation.height - 
		trend->plot_top_offset - trend->plot_buttom_offset - 
		trend->trend_top_margine - trend->trend_buttom_margine - 
		trend->main_label_area_height;
		
	
	trend->bar_area_height = trend->plot_area_height;
	trend->bar_area_width = trend->plot_area_width / BAR_ASPECT_RATIO;

	trend->bar_width = trend->bar_area_width / TREND_SIGNALS - 2 * trend->bar_margine;

	trend->plot_signals_area_width = trend->plot_area_width - trend->bar_area_width;

	trend->xc = trend->plot_area_width + trend->plot_left_offset + 
	   	trend->trend_left_margine - trend->bar_area_width; 
	trend->xend = trend->plot_left_offset + trend->trend_left_margine + TREND_DX;
	trend->yc = trend->plot_top_offset + trend->trend_top_margine + 
		trend->main_label_area_height + trend->plot_area_height;
}

