/******************************************************************************
 * Copyright 2004, 2005 Alex Karev.
 *
 * This file is part of gtkpivo library.
 *
 * gtkpivo library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation version 2.1 of the License.
 *
 * gtkpivo library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with gtkpivo library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ******************************************************************************/
/* vim: set ts=4 sw=4 : */
#ifndef __GTK_TREND_
#define __GTK_TREND_

#include <gtk/gtk.h>
#include <gdk/gdk.h>

/* Can't be more then 4, for now */
#define TREND_SIGNALS		4

#define TREND_HISTORY		512

#define GTK_TREND(obj)          GTK_CHECK_CAST (obj, gtk_trend_get_type (), GtkTrend)
#define GTK_TREND_CLASS(klass)  GTK_CHECK_CLASS_CAST (klass, gtk_trend_get_type (), GtkTrendClass)
#define GTK_IS_TREND(obj)       GTK_CHECK_TYPE (obj, gtk_trend_get_type ())

G_BEGIN_DECLS

typedef struct _GtkTrend        GtkTrend;
typedef struct _GtkTrendClass   GtkTrendClass;

struct _GtkTrend
{
	GtkWidget widget;
	
	GdkPixmap *bgpix;
	GdkGC *line_gc[ TREND_SIGNALS ];
	GdkGC *grid_line_gc1, *grid_line_gc2, *sel_line_gc;

	guint draw_value_selection_flag : 1;
	gint sel_x, sel_y;

	gint trend_left_margine,
		trend_right_margine,
		trend_top_margine,
		trend_buttom_margine;

	gint plot_top_offset,
		plot_buttom_offset,
		plot_left_offset,
		plot_right_offset;

	gint plot_area_width, 
		plot_area_height, 
		bar_area_height, 
		bar_area_width,
		plot_signals_area_width;

	gint main_label_area_height,
		main_label_area_width;

	gint values_area_height,
		values_area_width,
		values_area_margine,
		values_area_ystart,
		values_area_yend,
		values_area_xstart,
		values_area_xend;


	gint bar_margine, bar_width;
	
	/** Line star coordinates: x0 y0 */
	gint xc,xend,yc, button;


	/** History buffers */
	gint 
		head[ TREND_SIGNALS ], 
		n[ TREND_SIGNALS ],
		tail[ TREND_SIGNALS ];
	gdouble history[ TREND_SIGNALS ][ TREND_HISTORY ];

	GdkPoint points[TREND_HISTORY];
	

	GtkAdjustment *adj[ TREND_SIGNALS ];

	PangoFontDescription *font_grid_labels, *font_trend_label, *font_val;

	PangoLayout *layout;
  gchar *label;
};

struct _GtkTrendClass
{
	  GtkWidgetClass parent_class;
};


GtkWidget* gtk_trend_new( gchar *label,
	GtkAdjustment *adj1, GtkAdjustment *adj2,
	GtkAdjustment *adj3, GtkAdjustment *adj4	);
/* GtkWidget* gtk_trend_new(); */
GtkType gtk_trend_get_type(void);
void gtk_trend_update_plot(GtkTrend *t);
#if 0
void gtk_trend_append_values(GtkTrend *t, gfloat val1, gfloat val2, gfloat val3, gfloat val4);
#endif
GtkAdjustment*	gtk_trend_get_adjustment(GtkTrend *t, gint index);

G_END_DECLS

#endif

